/*
 * Girder 3.2 Plugin
 *
 * This plugin should support any HAUPPAUGE 878 or 88x based WinTV cards,
 * as well as WinTV-PVR USB, WinTV-PVR USB2, WinTV-PVR Pci, PVR250/350,
 * and WinTV-HD (assuming of course the particular model has the IR jack
 * on it)
 *
 * Copyright 2003 (C) Peter Heitman
 *
 */

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "HauppaugeIRRemote.h"
#include "irremote.h"
extern "C"
{
#include "girder.h"
}

// Local Defines

#define PLUGINCLASS     "HauppaugeIRRemote"
#define PLUGINNAME		"Hauppauge IRRemote"
#define PLUGINNUM		98

// Local Types

typedef struct
{
	int iKey;
	const char* pszString;
	
} KeyToStringMapping_t;


// Local Functions

LRESULT CALLBACK IRRemoteWndProc( HWND, UINT, WPARAM, LPARAM );
LPTSTR GetStringRes( int );
const char* lookupPVRKeyCode( int iKeyCode );
const char* lookupClassicKeyCode( int iKeyCode );


// Global Variables

static HANDLE    s_hDll;   // DLL's module handle
static HANDLE	 s_hWnd;   // Hidden Window handle
static int       s_iRepeatCount;
static bool      s_bActive = FALSE;

static s_functions SF;

void WINAPI gir_version( char* pszBuffer, int cbBuffer )
{
    strncpy( pszBuffer, "1.0", cbBuffer );
}

void WINAPI gir_name( char* pszBuffer, int cbBuffer )
{
    strncpy( pszBuffer, PLUGINNAME, cbBuffer );
}

void WINAPI gir_description( char* pszBuffer, int cbBuffer )
{
    strncpy( pszBuffer, "Hauppauge IRRemote Plugin written by Peter Heitman",
	     cbBuffer );
}

int WINAPI gir_devicenum()
{
    return PLUGINNUM;
}

int WINAPI gir_requested_api( int maxapi )
{
    return 1;
}

int WINAPI gir_open( int gir_major_ver, int gir_minor_ver,
		     int gir_micro_ver, p_functions p )
{
    if( p->size != sizeof( s_functions ) )
    {
	return GIR_FALSE;
    }
    
    SF = *p;

    WNDCLASS wc;
    
    wc.style         = CS_GLOBALCLASS;
    wc.lpfnWndProc   = (WNDPROC) IRRemoteWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = s_hDll;
    wc.hIcon         = NULL;
    wc.hCursor       = NULL;
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = (LPSTR )NULL;
    wc.lpszClassName = (LPSTR )PLUGINCLASS;
    
    if( !RegisterClass( &wc ) )
    {
	MessageBox( NULL,
		    GetStringRes( IDS_REGCLASSFAIL ),
		    (LPCTSTR )"HauppaugeIRRemote.dll",
		    MB_OK | MB_ICONEXCLAMATION );
	
	return FALSE;
    }
    
    s_hWnd = CreateWindow( (LPSTR )PLUGINCLASS, PLUGINNAME,
			   0, 0, 0, 0, 0, 0, 0, s_hDll, 0 );
    
    if( s_hWnd == 0 )
    {
	MessageBox( NULL,
		    GetStringRes( IDS_CREATEWINDOWFAIL ),
		    (LPCTSTR )"HauppaugeIRRemote.dll",
		    MB_OK | MB_ICONEXCLAMATION );
	
	return FALSE;
    }
    
    if( !IR_Open( s_hWnd, 0, FALSE, 0 ) )
    {
	MessageBox( NULL,
		    GetStringRes( IDS_OPENIRFAIL ),
		    (LPCTSTR )"HauppaugeIRRemote.dll",
		    MB_OK | MB_ICONEXCLAMATION );
	
	return FALSE;
    }

    return GIR_TRUE;
}

int WINAPI gir_close()
{
    KillTimer( s_hWnd, 1 );
    
    IR_Close( 0,0 );
    
    DestroyWindow( s_hWnd );
    
    s_hWnd = 0;
    
    if( !UnregisterClass( (LPSTR )PLUGINCLASS, s_hDll ) )
    {
	MessageBox( NULL,
		    GetStringRes( IDS_UNREGFAIL ),
		    (LPCTSTR )"HauppaugeIRRemote.dll",
		    MB_OK | MB_ICONEXCLAMATION );
	
	return FALSE;
    }

    return GIR_TRUE;
}

int WINAPI gir_start()
{
    s_bActive = TRUE;

    return GIR_TRUE;
}

int WINAPI gir_stop()
{
    s_bActive = FALSE;

    return GIR_TRUE;
}

BOOL WINAPI DllMain( HANDLE hDLL, DWORD dwReason, LPVOID lpReserved )
{
    switch( dwReason ) 
    {
    case DLL_PROCESS_ATTACH:
	    s_hDll = hDLL;
	    break;
    case DLL_THREAD_ATTACH:
	    break;
    case DLL_THREAD_DETACH:
	    break;
    case DLL_PROCESS_DETACH:
	    break;
    }
    
    return TRUE;
}


/****************************************************************************\
*
*  FUNCTION:    IRRemoteWndProc (standard window procedure INPUTS/RETURNS)
*
*  COMMENTS:    This is the window procedure for our custom control.
*
\****************************************************************************/

LRESULT CALLBACK IRRemoteWndProc( HWND hwnd, UINT msg,
				  WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
    case WM_TIMER:
    {
	if( s_bActive == FALSE ) break;
	
	WORD wKeyCode = IR_GetKeyCode( &s_iRepeatCount );
	
	if( wKeyCode != IR_NOKEY )
	{
	    int iRemote = ( wKeyCode >> 6 ) & 0x1f;
	    
	    int iKeyCode = ( wKeyCode & 0x3f );
	    
	    const char* pszString = 0;
	    
	    if( iRemote == 31 )
	    {
		pszString = lookupPVRKeyCode( iKeyCode );
	    }
	    else
	    {
		pszString = lookupClassicKeyCode( iKeyCode );
	    }
	    
	    char szBuffer[ 16 ];
	    
	    if( pszString == 0 )
	    {
		sprintf( szBuffer, "CODE%2x", iKeyCode );
		
		pszString = szBuffer;
	    }
	    
	    SF.send_event( pszString, 0, 0, PLUGINNUM );
	}
	
	break;
    }
    
    default:
	    
	return DefWindowProc( hwnd, msg, wParam, lParam );
    }
    
    return (LONG )TRUE;
}



/****************************************************************************\
*
*  FUNCTION:    GetStringRes( int id INPUT ONLY )
*
*  COMMENTS:    Load the resource string with the ID given, and return a
*               pointer to it.  Notice that the buffer is common memory so
*               the string must be used before this call is made a second time.
*
\****************************************************************************/

LPTSTR GetStringRes( int id )
{
    static TCHAR buffer[ MAX_PATH ];
    
    buffer[0]=0;
    
    LoadString( GetModuleHandle( NULL ), id, buffer, MAX_PATH );
    
    return buffer;
}


const char* lookupClassicKeyCode( int iKeyCode )
{
    const char* pszString = 0;
    
    switch( iKeyCode )
    {
    case 0: pszString = "0"; break;
    case 1: pszString = "1"; break;
    case 2: pszString = "2"; break;
    case 3: pszString = "3"; break;
    case 4: pszString = "4"; break;
    case 5: pszString = "5"; break;
    case 6: pszString = "6"; break;
    case 7: pszString = "7"; break;
    case 8: pszString = "8"; break;
    case 9: pszString = "9"; break;
    case 12: pszString = "RADIO"; break;
    case 13: pszString = "MUTE"; break;
    case 15: pszString = "TV"; break;
    case 16: pszString = "VOLUP"; break;
    case 17: pszString = "VOLDOWN"; break;
    case 30: pszString = "RESERVED"; break;
    case 32: pszString = "CHNLUP"; break;
    case 33: pszString = "CHNLDOWN"; break;
    case 34: pszString = "SOURCE"; break;
    case 38: pszString = "MINIMIZE"; break;
    case 46: pszString = "FULLSCREEN"; break;
    }
    
    return pszString;
}

const char* lookupPVRKeyCode( int iKeyCode )
{
    const char* pszString = 0;
    
    switch( iKeyCode )
    {
    case 0: pszString = "0"; break;
    case 1: pszString = "1"; break;
    case 2: pszString = "2"; break;
    case 3: pszString = "3"; break;
    case 4: pszString = "4"; break;
    case 5: pszString = "5"; break;
    case 6: pszString = "6"; break;
    case 7: pszString = "7"; break;
    case 8: pszString = "8"; break;
    case 9: pszString = "9"; break;
    case 46: pszString = "GREEN"; break;
    case 56: pszString = "YELLOW"; break;
    case 11: pszString = "RED"; break;
    case 41: pszString = "BLUE"; break;
    case 12: pszString = "FUNC"; break;
    case 13: pszString = "MENU"; break;
    case 15: pszString = "MUTE"; break;
    case 16: pszString = "VOLUP"; break;
    case 17: pszString = "VOLDOWN"; break;
    case 32: pszString = "CHNLUP"; break;
    case 33: pszString = "CHNLDOWN"; break;
    case 61: pszString = "GRNPOWER"; break;
    case 31: pszString = "BACK"; break;
    case 37: pszString = "OK"; break;
    case 59: pszString = "GO"; break;
    case 60: pszString = "FULLSCREEN"; break;
    case 55: pszString = "REC"; break;
    case 54: pszString = "STOP"; break;
    case 48: pszString = "PAUSE"; break;
    case 53: pszString = "PLAY"; break;
    case 50: pszString = "REWIND"; break;
    case 52: pszString = "FASTFWD"; break;
    case 30: pszString = "SKIPFWD"; break;
    case 36: pszString = "SKIPREV"; break;
    }
    
    return pszString;
}


