{
WTVIRBridge
Copyright (c) 2004-07 Oliver Frietsch

WTVIRBridge is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

WTVIRBridge is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
}

unit fakewinlircthread;

interface

{$INCLUDE defines.inc}

{$IFDEF WINLIRC}
  uses
    Windows, Classes, BlckSock, WinSock, SyncObjs, Contnrs, SysUtils, CoolTrayIcon;

  type
    TFakeWinLIRCThread = class(TThread)
    private
      server: TTCPBlockSocket;
      connections: TObjectList;
      keyqueuelock: TCriticalSection;
      keyqueue: TStringList;
      lastaddress, lastkeycode, repeatcount: Integer;
    protected
      procedure Execute; override;
    public
      constructor Create;
      destructor Destroy; override;
      procedure SendKey (Address, KeyCode: Integer);
    end;
{$ENDIF}

implementation

{$IFDEF WINLIRC}
  uses
    main, lngstrings, log;

  constructor TFakeWinLIRCThread.Create;
  begin
    inherited Create (True);
    server := TTCPBlockSocket.Create;
    connections := TObjectList.Create (False);
    keyqueuelock := TCriticalSection.Create;
    keyqueue := TStringList.Create;
    lastaddress := -1;
    lastkeycode := -1;
    repeatcount := 0;
    server.Bind ('127.0.0.1', '8765');
    if server.LastError <> 0
      then begin
             WTVIRBridgeMainForm.ShowBalloonHint (ErrorTitle, ErrorBindServer, bitError, 20);
             server.Free;
             server := nil;
             LogForm.LogEvent (BindServerFailed);
           end
      else begin
             server.Listen;
             LogForm.LogEvent (BindServerSuccess);
             Resume;
           end;
  end;

  destructor TFakeWinLIRCThread.Destroy;
  begin
    keyqueue.Free;
    keyqueuelock.Free;
    connections.Free;
    server.Free;
    inherited Destroy;
  end;

  procedure TFakeWinLIRCThread.Execute;
  var
    client: TTCPBlockSocket;
    sendstr: String;
    index: Integer;
  begin
    repeat
      if server.CanRead (0)
        then begin
               client := TTCPBlockSocket.Create;
               client.Socket := server.Accept;
               connections.Add (client);
             end;
      sendstr := '';
      keyqueuelock.Acquire;
      for index := 0 to keyqueue.Count-1 do
        sendstr := sendstr + keyqueue.Strings[index] + LF;
      keyqueue.Clear;
      keyqueuelock.Release;
      index := 0;
      while index < connections.Count do
        begin
          client := connections[index] as TTCPBlockSocket;
          client.Purge;
          if sendstr <> ''
            then client.SendString (sendstr);
          if client.LastError <> 0
            then begin
                   client.Free;
                   connections.Delete (index);
                   if connections.Capacity >= 2*connections.Count
                     then connections.Capacity := connections.Count;
                 end
            else Inc (index);
        end;
      Sleep (50);
    until Terminated;
  end;

  procedure TFakeWinLIRCThread.SendKey (Address, KeyCode: Integer);

    function PrependZeros (Text: String; RequiredLength: Integer): String;
    begin
      while Length (Text) < RequiredLength do
        Text := '0' + Text;
      Result := Text;
    end;

    function KillSpaces (Text: String): String;
    var
      index: Integer;
    begin
      index := 1;
      while index <= Length (Text) do
        if Text[index] = ' '
          then Delete (Text, index, 1)
          else Inc (index);
      Result := Text;
    end;

    function NumberToHex (Number: Byte): String;
      function HexSymbol (Number: Byte): Char;
      //Expects a Number between 0 and 15!!!
      begin
        if Number < 10
          then Result := Chr (Ord ('0') + Number)
          else Result := Chr (Ord ('A') - 10 + Number);
      end;
    begin
      Result := HexSymbol (Number div 16) + HexSymbol (Number mod 16);
    end;

  begin
    if server <> nil
      then begin
             if (Address = lastaddress) and (KeyCode = lastkeycode)
               then Inc (repeatcount)
               else repeatcount := 0;
             if repeatcount > 255 then
               repeatcount := 0;
             lastaddress := Address;
             lastkeycode := KeyCode;
             keyqueuelock.Acquire;
             keyqueue.Add (
                           PrependZeros (IntToStr (Address), 8) + PrependZeros (IntToStr (KeyCode), 8) + ' ' +
                           NumberToHex (repeatcount) + ' ' +
                           KillSpaces (GetKeyName (Address, KeyCode)) + ' ' +
                           KillSpaces (GetRemoteName (Address))
                          );
             keyqueuelock.Release;
           end;
  end;
{$ENDIF}

end.
