/*
 * This file was produced from documentation provided by
 * Hauppauge Computer Works. It is released under the
 * GNU General Public License
 *
 * Peter Heitman, Oct 2003
 */
 
#ifndef _IRREMOTE_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * General Notes
 *
 * The basic sequence is IR_Open(), and this starts a timer;
 *  poll using IR_GetKeyCode() called from the WM_TIMER messages,
 *  and then call KillTimer( hWnd, 1) and IR_Close() when you are done.
 */

/*
 * Defines
 */

#define IR_NOKEY	0x1fff
 
/*
 * IR_Open
 *
 * Parameters:
 *
 *   hWnd is the window handle of your app. Timer events WILL be sent to
 *     this window if IR_Open is successful; the interval is nominally 85ms
 *
 *   msg is UNUSED, set it to 0
 * 
 *   Verbose is a flag to indicate if a message box should be displayed
 *     with status during the open
 *  
 *   myIRPort is the I2C address to connect to; by default it should be 0,
 *     which does a search for any suitable device. You should set it to 0
 *     unless you have a good reason not to (hint: you don't :-)
 *
 * Return Value:
 *
 *   returns non-0 if it was succesful
 */

BOOL WINAPI IR_Open( HWND hWnd, UINT msg, BOOL Verbose, WORD myIRPort );

/*
 * IR_GetVersion
 *
 * Return Value:
 *
 *   returns the FW version of the found IR controller in the lower 6 bits; 
 *     only useful for testing
 */
 
DWORD WINAPI IR_GetVersion();

/*
 * IR_Close
 *
 * Closes down the existing connections to the IR device
 *   (Does NOT stop the TIMER; do that yourself via KillTimer( hWnd, 1))
 *
 * Parameters:
 *
 *   hWnd - currently NOT USED, set to 0
 *
 *   msg - currently NOT USED, set to 0
 *
 * Return Value:
 *
 *   returns TRUE
 */

BOOL WINAPI IR_Close( HWND hWnd, UINT msg );

/*
 * IR_GetKeyCode
 *
 * Parameters:
 *
 *   piRepeatCount  points to a value that contains the current repeat count; 
 *     used to do auto repeat. This pointed-to value SHOULD be 0 initially;
 *     and should probably never be touched; and it MUST be Global/static
 *     (not stack based) (the IR_GetKeyCode function assumes this state
 *     variable is preserved across calls)
 *
 * Return Value:
 *
 *   returns the current IR button press, or IR_NOKEY for no button pressed
 *     since last poll
 *
 *   The format of the returned KeyCode is the 14bit RC5 code; 
 *   bit:   13 12 11  10 9  8  7  6  5  4  3  2  1  0
 *   cmd:    S  S  T  A4 A3 A2 A1 A0 C5 C4 C3 C2 C1 C0
 *
 *   S= status, always 1
 *   T= toggle, changes for every physical keypress (but not auto repeats
 *         from held keys)
 *   Ax = Address bits (0=older black remote, 31=newer PVR remote)
 *   Cx = Key codes (look at the full list for each remote in
 *                   \windows\IRRemote.ini file)
 */
 
WORD WINAPI IR_GetKeyCode( int* piRepeatCount );

#ifdef __cplusplus
}
#endif

#endif
