{
WTVIRBridge
Copyright (c) 2004-07 Oliver Frietsch

WTVIRBridge is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

WTVIRBridge is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
}

unit main; {OK}

interface

uses
  Windows, Messages, Forms, Menus, ExtCtrls, ImgList, Controls, CoolTrayIcon,
  StdCtrls, Classes, Math, Registry, SysUtils, AppEvnts, PJVersionInfo;

type
  TWTVIRBridgeMainForm = class(TForm)
    TrayIcon: TCoolTrayIcon;
    TrayIconList: TImageList;
    BalloonIconList: TImageList;
    BalloonShape: TShape;
    BalloonIconImg: TImage;
    BalloonTitleLbl: TLabel;
    BalloonTextLbl: TLabel;
    BalloonTimer: TTimer;
    TrayMenu: TPopupMenu;
    AboutItem: TMenuItem;
    ExitItem: TMenuItem;
    LogItem: TMenuItem;
    Separator1Item: TMenuItem;
    UserStartupItem: TMenuItem;
    Separator2Item: TMenuItem;
    ApplicationEvents: TApplicationEvents;
    ReOpenItem: TMenuItem;
    Separator3Item: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure BalloonTimerTimer(Sender: TObject);
    procedure BalloonShapeMouseDown(Sender: TObject; Button: TMouseButton;
                                    Shift: TShiftState; X, Y: Integer);
    procedure ExitItemClick(Sender: TObject);
    procedure LogItemClick(Sender: TObject);
    procedure AboutItemClick(Sender: TObject);
    procedure UserStartupItemClick(Sender: TObject);
    procedure ApplicationEventsRestore(Sender: TObject);
    procedure ReOpenItemClick(Sender: TObject);
  private
    FRepeatCount: Integer;
    FTimersBeforeIconReset: Byte;
    FIROpened: Boolean;
    procedure WMTimer (var Msg: TMessage); message WM_Timer;
    procedure WMUser (var Msg: TMessage); message wm_User;
    procedure StartupIR;
    procedure ShutdownIR;
  public
    procedure ShowBalloonHint (Title: String; Text: String; IconType: TBalloonHintIcon;
                               TimeoutSecs: TBalloonHintTimeOut);
    function CheckUserStartup: Boolean;
  end;

procedure TaskbarButton (Visible: Boolean; IgnoreForm: TForm);
procedure ShowBalloonHint (Title: String; Text: String; IconType: TBalloonHintIcon;
                           TimeoutSecs: TBalloonHintTimeOut);
function GetVersion: String;
function GetCopyright: String;

var
  WTVIRBridgeMainForm: TWTVIRBridgeMainForm;
  Parameters: record
                NoBalloon: Boolean;
              end;  

implementation

uses
  about, broadcast, irremote, lngstrings, log;

{$R *.DFM}

procedure TWTVIRBridgeMainForm.FormCreate(Sender: TObject);
begin
  LogForm := TLogForm.Create (Self);
  ClientWidth := 0;
  ClientHeight := 0;
  TrayIcon.IconVisible := True;
  UserStartupItem.Checked := CheckUserStartup;
  FRepeatCount := 0;
  FTimersBeforeIconReset := 0;
  FIROpened := False;
  BeginBroadcast;
  StartupIR;
end;

procedure TWTVIRBridgeMainForm.WMTimer (var Msg: TMessage);
var
  address, keycode: Integer;
begin
  if Msg.WParam = 1
    then begin
           if FTimersBeforeIconReset > 0
             then Dec (FTimersBeforeIconReset);
           if IR_GetSystemKeyCode (@FRepeatCount, @address, @keycode)
             then begin
                    if TrayIcon.IconIndex <> 1
                      then TrayIcon.IconIndex := 1;
                    FTimersBeforeIconReset := 3; {*85 ms}
                    BroadcastKey (address, keycode);
                    LogForm.LogKey (address, keycode);
                  end;
           if (FTimersBeforeIconReset = 0) and (TrayIcon.IconIndex <> 0)
             then TrayIcon.IconIndex := 0;
         end;
  Msg.Result := 0;
end;

procedure TWTVIRBridgeMainForm.WMUser (var Msg: TMessage);
begin
  if Msg.WParam >= 0
    then begin
           LogForm.LogEvent(IROpenRequest);
           StartupIR; // includes "ReStartupIR"
         end
    else begin
           LogForm.LogEvent(IRCloseRequest);
           ShutdownIR;
         end;
end;

procedure TWTVIRBridgeMainForm.FormDestroy(Sender: TObject);
begin
  ShutdownIR;
  EndBroadcast;
  TrayIcon.IconVisible := False;
  LogForm.Free;
end;

procedure TWTVIRBridgeMainForm.ShowBalloonHint (Title: String; Text: String;
                                     IconType: TBalloonHintIcon;
                                     TimeoutSecs: TBalloonHintTimeOut);
var
  workarearect: TRect;
  version: TOSVersionInfo;
  nativeballoon: Boolean;
begin
  if Parameters.NoBalloon
    then Exit;
  version.dwOSVersionInfoSize := SizeOf (version);
  nativeballoon := GetVersionEx (version);
  nativeballoon := nativeballoon and
                   (version.dwPlatformId = VER_PLATFORM_WIN32_NT) and
                   (version.dwMajorVersion >= 5); //Windows 2000, XP or newer
  if nativeballoon
    then nativeballoon := TrayIcon.ShowBalloonHint (Title, Text, IconType, TimeoutSecs);
  if not nativeballoon
    then begin
           BalloonTitleLbl.Caption := Title;
           BalloonTextLbl.Caption := Text;
           BalloonIconList.GetIcon (Ord (IconType) - 1, BalloonIconImg.Picture.Icon);
           BalloonTimer.Interval := TimeoutSecs * 1000;
           ClientWidth := Max (BalloonTextLbl.Left*2 + BalloonTextLbl.Width,
                               BalloonTitleLbl.Left + BalloonTextLbl.Left + BalloonTitleLbl.Width);
           ClientHeight := BalloonTextLbl.Top + BalloonTextLbl.Height + BalloonIconImg.Top;
           if not SystemParametersInfo (spi_GetWorkArea, 0, @workarearect, 0)
             then begin
                    workarearect.Top := 0;
                    workarearect.Left := 0;
                    workarearect.Bottom := Screen.Height-1;
                    workarearect.Right := Screen.Width-1;
                  end;
           Left := workarearect.Right - Width - 8;
           Top := workarearect.Bottom - Height - 8;
           BalloonTimer.Enabled := True;
           Show;
         end;
end;

procedure TWTVIRBridgeMainForm.BalloonTimerTimer(Sender: TObject);
begin
  BalloonTimer.Enabled := False;
  Hide;
end;

procedure TWTVIRBridgeMainForm.BalloonShapeMouseDown(Sender: TObject; Button: TMouseButton;
                                          Shift: TShiftState; X, Y: Integer);
begin
  BalloonTimer.OnTimer (Self);
end;

procedure TWTVIRBridgeMainForm.ExitItemClick(Sender: TObject);
begin
  Close;
end;

procedure TWTVIRBridgeMainForm.LogItemClick(Sender: TObject);
begin
  LogForm.Show;
end;

procedure TWTVIRBridgeMainForm.AboutItemClick(Sender: TObject);
begin
  AboutForm.Show;
end;

procedure TaskbarButton (Visible: Boolean; IgnoreForm: TForm);
var
  index: Integer;
begin
  if Visible
    then begin
           SetWindowLong (Application.Handle, gwl_ExStyle,
                          GetWindowLong (Application.Handle, gwl_ExStyle) and not ws_ex_ToolWindow);
           SetForegroundWindow (Application.Handle);
         end
    else begin
           for index := 0 to Application.ComponentCount-1 do
             if (Application.Components[index] is TForm) and
                (Application.Components[index] as TForm).Visible and
                (Application.Components[index] <> IgnoreForm)
               then Exit;
           SetWindowLong (Application.Handle, gwl_ExStyle,
                          GetWindowLong (Application.Handle, gwl_ExStyle) or ws_ex_ToolWindow);
         end;
end;

procedure ShowBalloonHint (Title: String; Text: String; IconType: TBalloonHintIcon;
                           TimeoutSecs: TBalloonHintTimeOut);
begin
  WTVIRBridgeMainForm.ShowBalloonHint (Title, Text, IconType, TimeoutSecs);
end;

function TWTVIRBridgeMainForm.CheckUserStartup: Boolean;
var
  reg: TRegistry;
begin
  Result := False;
  reg := nil;
  try
    reg := TRegistry.Create;
    if reg.OpenKey ('\Software\Microsoft\Windows\CurrentVersion\Run', True)
      then begin
             Result := reg.ValueExists ('WTVIRBridge') and
                       (LowerCase (reg.ReadString ('WTVIRBridge')) = LowerCase ('"' + Application.ExeName + '"'));
             reg.CloseKey;
           end;
  finally
    reg.Free;
  end;
end;

procedure TWTVIRBridgeMainForm.UserStartupItemClick(Sender: TObject);
var
  reg: TRegistry;
begin
  UserStartupItem.Checked := not UserStartupItem.Checked;
  if UserStartupItem.Checked
    then begin
           reg := nil;
           try
             reg := TRegistry.Create;
             if reg.OpenKey ('\Software\Microsoft\Windows\CurrentVersion\Run', True)
               then begin
                      reg.WriteString ('WTVIRBridge', '"' + Application.ExeName  + '"');
                      reg.CloseKey;
                    end;
           finally
             reg.Free;
           end;
         end
    else begin
           reg := nil;
           try
             reg := TRegistry.Create;
             if reg.OpenKey ('\Software\Microsoft\Windows\CurrentVersion\Run', True)
               then begin
                      if reg.ValueExists ('WTVIRBridge')
                        then reg.DeleteValue ('WTVIRBridge');
                      reg.CloseKey;
                    end;
           finally
             reg.Free;
           end;
         end;
end;

procedure TWTVIRBridgeMainForm.ApplicationEventsRestore(Sender: TObject);
begin
  Hide;
end;

procedure TWTVIRBridgeMainForm.StartupIR;
begin
  ShutdownIR; //Leads to clean re-open if already open
  TrayIcon.IconIndex := 2;
  FIROpened := IR_Open (Handle, 0, False, 0);
  if FIROpened
    then begin
           TrayIcon.IconIndex := 0;
           LogForm.LogEvent(IROpenSuccess);
         end
    else begin
           TrayIcon.IconIndex := 3;
           ShowBalloonHint (ErrorTitle, ErrorOpenDevice, bitError, 20);
           LogForm.LogEvent(IROpenFailed);
         end;
end;

procedure TWTVIRBridgeMainForm.ShutdownIR;
begin
  if FIROpened
    then begin
           KillTimer (Handle, 1);
           FIROpened := not IR_Close (0, 0);
           if not FIROpened
             then LogForm.LogEvent(IRCloseSuccess)
             else LogForm.LogEvent(IRCloseFailed);
           TrayIcon.IconIndex := 3;
         end;
end;

function GetVersion: String;
var
  vi: TPJVersionInfo;
begin
  vi := TPJVersionInfo.Create (nil);
  vi.FileName := ''; //Read from this EXE
  Result := vi.FileVersion;
  vi.Free;
end;

function GetCopyright: String;
var
  vi: TPJVersionInfo;
begin
  vi := TPJVersionInfo.Create (nil);
  vi.FileName := ''; //Read from this EXE
  Result := vi.LegalCopyright;
  vi.Free;
end;

procedure TWTVIRBridgeMainForm.ReOpenItemClick(Sender: TObject);
begin
  StartupIR;
end;

end.
